@@ -1,22 +1,26 @@
 /**
  * Object oriented C module to send ICMP and ICMPv6 `echo's.
- * Copyright (C) 2006-2009  Florian octo Forster <octo at verplant.org>
+ * Copyright (C) 2006-2011  Florian octo Forster <ff at octo.it>
  *
- * This program is free software; you can redistribute it and/or modify
- * it under the terms of the GNU General Public License as published by
- * the Free Software Foundation; only version 2 of the License is
- * applicable.
- *
- * This program is distributed in the hope that it will be useful,
- * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- * GNU General Public License for more details.
- *
- * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA
+ * This library is free software; you can redistribute it and/or modify it
+ * under the terms of the GNU Lesser General Public License as published by the
+ * Free Software Foundation; either version 2.1 of the License, or (at your
+ * option) any later version.
+ * 
+ * This library is distributed in the hope that it will be useful, but WITHOUT
+ * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
+ * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License
+ * for more details.
+ * 
+ * You should have received a copy of the GNU Lesser General Public License
+ * along with this library; if not, write to the Free Software Foundation,
+ * Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
  */
 
+#ifdef __APPLE__
+#define __APPLE_USE_RFC_3542
+#endif
+
 #if HAVE_CONFIG_H
 # include <config.h>
 #endif
@@ -117,6 +121,7 @@ struct pinghost
 	double                   latency;
 	uint32_t                 dropped;
 	int                      recv_ttl;
+	uint8_t                  recv_qos;
 	char                    *data;
 
 	void                    *context;
@@ -129,6 +134,7 @@ struct pingobj
 	double                   timeout;
 	int                      ttl;
 	int                      addrfamily;
+	uint8_t                  qos;
 	char                    *data;
 
 	struct sockaddr         *srcaddr;
@@ -350,9 +356,10 @@ static pinghost_t *ping_receive_ipv4 (pingobj_t *obj, char *buffer,
 				ident, seq);
 	}
 
-	if (ptr != NULL)
-		ptr->recv_ttl = ip_hdr->ip_ttl;
-
+	if (ptr != NULL){
+		ptr->recv_ttl = (int)     ip_hdr->ip_ttl;
+		ptr->recv_qos = (uint8_t) ip_hdr->ip_tos;
+	}
 	return (ptr);
 }
 
@@ -447,9 +454,10 @@ static int ping_receive_one (pingobj_t *obj, const pinghost_t *ph,
 	 * reply. The right object will be returned by ping_receive_ipv*(). For
 	 * now, we can only rely on ph->fd and ph->addrfamily. */
 
-	struct timeval diff;
+	struct timeval diff, pkt_now = *now;
 	pinghost_t *host = NULL;
 	int recv_ttl;
+	uint8_t recv_qos;
 	
 	/*
 	 * Set up the receive buffer..
@@ -495,15 +503,29 @@ static int ping_receive_one (pingobj_t *obj, const pinghost_t *ph,
 
 	/* Iterate over all auxiliary data in msghdr */
 	recv_ttl = -1;
+	recv_qos = 0;
 	for (cmsg = CMSG_FIRSTHDR (&msghdr); /* {{{ */
 			cmsg != NULL;
 			cmsg = CMSG_NXTHDR (&msghdr, cmsg))
 	{
-		if (ph->addrfamily == AF_INET) /* {{{ */
+		if (cmsg->cmsg_level == SOL_SOCKET)
+		{
+#ifdef SO_TIMESTAMP
+			if (cmsg->cmsg_type == SO_TIMESTAMP)
+				memcpy (&pkt_now, CMSG_DATA (cmsg), sizeof (pkt_now));
+#endif /* SO_TIMESTAMP */
+		}
+		else if (ph->addrfamily == AF_INET) /* {{{ */
 		{
 			if (cmsg->cmsg_level != IPPROTO_IP)
 				continue;
 
+			if (cmsg->cmsg_type == IP_TOS)
+			{
+				memcpy (&recv_qos, CMSG_DATA (cmsg),
+						sizeof (recv_qos));
+				dprintf ("TOSv4 = 0x%02"PRIx8";\n", recv_qos);
+			} else
 			if (cmsg->cmsg_type == IP_TTL)
 			{
 				memcpy (&recv_ttl, CMSG_DATA (cmsg),
@@ -521,6 +543,13 @@ static int ping_receive_one (pingobj_t *obj, const pinghost_t *ph,
 			if (cmsg->cmsg_level != IPPROTO_IPV6)
 				continue;
 
+			if (cmsg->cmsg_type == IPV6_TCLASS)
+			{
+				memcpy (&recv_qos, CMSG_DATA (cmsg),
+						sizeof (recv_qos));
+				dprintf ("TOSv6 = 0x%02"PRIx8";\n", recv_qos);
+			} else
+#ifdef IPV6_HOPLIMIT
 			if (cmsg->cmsg_type == IPV6_HOPLIMIT)
 			{
 				memcpy (&recv_ttl, CMSG_DATA (cmsg),
@@ -528,6 +557,25 @@ static int ping_receive_one (pingobj_t *obj, const pinghost_t *ph,
 				dprintf ("TTLv6 = %i;\n", recv_ttl);
 			}
 			else
+#endif
+#ifdef IPV6_UNICAST_HOPS
+			if (cmsg->cmsg_type == IPV6_UNICAST_HOPS)
+			{
+				memcpy (&recv_ttl, CMSG_DATA (cmsg),
+						sizeof (recv_ttl));
+				dprintf ("TTLv6 = %i;\n", recv_ttl);
+			}
+			else
+#endif
+#ifdef IPV6_MULTICAST_HOPS
+			if (cmsg->cmsg_type == IPV6_MULTICAST_HOPS)
+			{
+				memcpy (&recv_ttl, CMSG_DATA (cmsg),
+						sizeof (recv_ttl));
+				dprintf ("TTLv6 = %i;\n", recv_ttl);
+			}
+			else
+#endif
 			{
 				dprintf ("Not handling option %i.\n",
 						cmsg->cmsg_type);
@@ -561,13 +609,13 @@ static int ping_receive_one (pingobj_t *obj, const pinghost_t *ph,
 	}
 
 	dprintf ("rcvd: %12i.%06i\n",
-			(int) now->tv_sec,
-			(int) now->tv_usec);
+			(int) pkt_now.tv_sec,
+			(int) pkt_now.tv_usec);
 	dprintf ("sent: %12i.%06i\n",
 			(int) host->timer->tv_sec,
 			(int) host->timer->tv_usec);
 
-	if (ping_timeval_sub (now, host->timer, &diff) < 0)
+	if (ping_timeval_sub (&pkt_now, host->timer, &diff) < 0)
 	{
 		timerclear (host->timer);
 		return (-1);
@@ -579,6 +627,7 @@ static int ping_receive_one (pingobj_t *obj, const pinghost_t *ph,
 
 	if (recv_ttl >= 0)
 		host->recv_ttl = recv_ttl;
+	host->recv_qos = recv_qos;
 
 	host->latency  = ((double) diff.tv_usec) / 1000.0;
 	host->latency += ((double) diff.tv_sec)  * 1000.0;
@@ -931,6 +980,50 @@ static int ping_set_ttl (pinghost_t *ph, int ttl)
 	return (ret);
 }
 
+/*
+ * Set the TOS of a socket protocol independently.
+ *
+ * Using SOL_SOCKET / SO_PRIORITY might be a protocol independent way to
+ * set this. See socket(7) for details.
+ */
+static int ping_set_qos (pingobj_t *obj, pinghost_t *ph, uint8_t qos)
+{
+	int ret = EINVAL;
+	char errbuf[PING_ERRMSG_LEN];
+
+	if (ph->addrfamily == AF_INET)
+	{
+		dprintf ("Setting TP_TOS to %#04"PRIx8"\n", qos);
+		ret = setsockopt (ph->fd, IPPROTO_IP, IP_TOS,
+				&qos, sizeof (qos));
+		if (ret != 0)
+		{
+			ret = errno;
+			ping_set_error (obj, "ping_set_qos",
+					sstrerror (ret, errbuf, sizeof (errbuf)));
+			dprintf ("Setting TP_TOS failed: %s\n", errbuf);
+		}
+	}
+	else if (ph->addrfamily == AF_INET6)
+	{
+		/* IPV6_TCLASS requires an "int". */
+		int tmp = (int) qos;
+
+		dprintf ("Setting IPV6_TCLASS to %#04"PRIx8" (%i)\n", qos, tmp);
+		ret = setsockopt (ph->fd, IPPROTO_IPV6, IPV6_TCLASS,
+				&tmp, sizeof (tmp));
+		if (ret != 0)
+		{
+			ret = errno;
+			ping_set_error (obj, "ping_set_qos",
+					sstrerror (ret, errbuf, sizeof (errbuf)));
+			dprintf ("Setting IPV6_TCLASS failed: %s\n", errbuf);
+		}
+	}
+
+	return (ret);
+}
+
 static int ping_get_ident (void)
 {
 	int fd;
@@ -1030,12 +1123,13 @@ pingobj_t *ping_construct (void)
 
 	if ((obj = (pingobj_t *) malloc (sizeof (pingobj_t))) == NULL)
 		return (NULL);
-	memset (obj, '\0', sizeof (pingobj_t));
+	memset (obj, 0, sizeof (pingobj_t));
 
 	obj->timeout    = PING_DEF_TIMEOUT;
 	obj->ttl        = PING_DEF_TTL;
 	obj->addrfamily = PING_DEF_AF;
 	obj->data       = strdup (PING_DEF_DATA);
+	obj->qos        = 0;
 
 	return (obj);
 }
@@ -1081,6 +1175,16 @@ int ping_setopt (pingobj_t *obj, int option, void *value)
 
 	switch (option)
 	{
+		case PING_OPT_QOS:
+		{
+			pinghost_t *ph;
+
+			obj->qos = *((uint8_t *) value);
+			for (ph = obj->head; ph != NULL; ph = ph->next)
+				ping_set_qos (obj, ph, obj->qos);
+			break;
+		}
+
 		case PING_OPT_TIMEOUT:
 			obj->timeout = *((double *) value);
 			if (obj->timeout < 0.0)
@@ -1349,7 +1453,7 @@ int ping_host_add (pingobj_t *obj, const char *host)
 			snprintf (errmsg, PING_ERRMSG_LEN, "Unknown `ai_family': %i", ai_ptr->ai_family);
 			errmsg[PING_ERRMSG_LEN - 1] = '\0';
 
-			dprintf (errmsg);
+			dprintf ("%s", errmsg);
 			ping_set_error (obj, "getaddrinfo", errmsg);
 			continue;
 		}
@@ -1395,7 +1499,7 @@ int ping_host_add (pingobj_t *obj, const char *host)
 			{
 #if WITH_DEBUG
 				char errbuf[PING_ERRMSG_LEN];
-				dprintf ("setsockopt: %s\n",
+				dprintf ("setsockopt (SO_BINDTODEVICE): %s\n",
 						sstrerror (errno, errbuf, sizeof (errbuf)));
 #endif
 				ping_set_errno (obj, errno);
@@ -1405,7 +1509,29 @@ int ping_host_add (pingobj_t *obj, const char *host)
 			}
 		}
 #endif /* SO_BINDTODEVICE */
+#ifdef SO_TIMESTAMP
+		if (1) /* {{{ */
+		{
+			int status;
+			int opt = 1;
 
+			status = setsockopt (ph->fd,
+					SOL_SOCKET, SO_TIMESTAMP,
+					&opt, sizeof (opt));
+			if (status != 0)
+			{
+#if WITH_DEBUG
+				char errbuf[PING_ERRMSG_LEN];
+				dprintf ("setsockopt (SO_TIMESTAMP): %s\n",
+						sstrerror (errno, errbuf, sizeof (errbuf)));
+#endif
+				ping_set_errno (obj, errno);
+				close (ph->fd);
+				ph->fd = -1;
+				continue;
+			}
+		} /* }}} if (1) */
+#endif /* SO_TIMESTAMP */
 		assert (sizeof (struct sockaddr_storage) >= ai_ptr->ai_addrlen);
 		memset (ph->addr, '\0', sizeof (struct sockaddr_storage));
 		memcpy (ph->addr, ai_ptr->ai_addr, ai_ptr->ai_addrlen);
@@ -1436,23 +1562,43 @@ int ping_host_add (pingobj_t *obj, const char *host)
 
 		if (ph->addrfamily == AF_INET)
 		{
-			int opt = 1;
+			int opt;
 
+#ifdef IP_RECVTOS
+			/* Enable receiving the TOS field */
+			opt = 1;
+			setsockopt (ph->fd, IPPROTO_IP, IP_RECVTOS,
+					&opt, sizeof (opt));
+#endif	/* IP_RECVTOS */
+
+			/* Enable receiving the TTL field */
+			opt = 1;
 			setsockopt (ph->fd, IPPROTO_IP, IP_RECVTTL,
 					&opt, sizeof (opt));
 		}
-#if defined(IPPROTO_IPV6) && defined(IPV6_RECVHOPLIMIT)
+#if defined(IPV6_RECVHOPLIMIT) || defined(IPV6_RECVTCLASS)
 		else if (ph->addrfamily == AF_INET6)
 		{
-			int opt = 1;
+			int opt;
 
+# if defined(IPV6_RECVHOPLIMIT)
+			/* For details see RFC 3542, section 6.3. */
+			opt = 1;
 			setsockopt (ph->fd, IPPROTO_IPV6, IPV6_RECVHOPLIMIT,
 					&opt, sizeof (opt));
+# endif /* IPV6_RECVHOPLIMIT */
+
+# if defined(IPV6_RECVTCLASS)
+			/* For details see RFC 3542, section 6.5. */
+			opt = 1;
+			setsockopt (ph->fd, IPPROTO_IPV6, IPV6_RECVTCLASS,
+					&opt, sizeof (opt));
+# endif /* IPV6_RECVTCLASS */
 		}
-#endif
+#endif /* IPV6_RECVHOPLIMIT || IPV6_RECVTCLASS */
 
 		break;
-	}
+	} /* for (ai_ptr = ai_list; ai_ptr != NULL; ai_ptr = ai_ptr->ai_next) */
 
 	freeaddrinfo (ai_list);
 
@@ -1484,6 +1630,7 @@ int ping_host_add (pingobj_t *obj, const char *host)
 	}
 
 	ping_set_ttl (ph, obj->ttl);
+	ping_set_qos (obj, ph, obj->qos);
 
 	return (0);
 } /* int ping_host_add */
@@ -1662,6 +1809,16 @@ int ping_iterator_get_info (pingobj_iter_t *iter, int info,
 			*((int *) buffer) = iter->recv_ttl;
 			ret = 0;
 			break;
+
+		case PING_INFO_RECV_QOS:
+			ret = ENOMEM;
+			if (*buffer_len>sizeof(unsigned)) *buffer_len=sizeof(unsigned);
+			if (!*buffer_len) *buffer_len=1;
+			if (orig_buffer_len < *buffer_len)
+				break;
+			memcpy(buffer,&iter->recv_qos,*buffer_len);
+			ret = 0;
+			break;
 	}
 
 	return (ret);
